@@ -1,5 +1,6 @@
+from __future__ import annotations
+
 from typing import Any
-from typing import cast
 from typing import Dict
 from typing import List
 from typing import Optional
@@ -23,11 +24,14 @@ from sqlalchemy.util import topological
 from ..util import exc
 from ..util.sqla_compat import _columns_for_constraint
 from ..util.sqla_compat import _copy
+from ..util.sqla_compat import _copy_expression
 from ..util.sqla_compat import _ensure_scope_for_ddl
 from ..util.sqla_compat import _fk_is_self_referential
+from ..util.sqla_compat import _idx_table_bound_expressions
 from ..util.sqla_compat import _insert_inline
 from ..util.sqla_compat import _is_type_bound
 from ..util.sqla_compat import _remove_column_from_collection
+from ..util.sqla_compat import _resolve_for_variant
 from ..util.sqla_compat import _select
 
 if TYPE_CHECKING:
@@ -121,7 +125,7 @@ class BatchOperationsImpl:
                         schema=self.schema,
                         autoload_with=self.operations.get_bind(),
                         *self.reflect_args,
-                        **self.reflect_kwargs
+                        **self.reflect_kwargs,
                     )
                     reflected = True
 
@@ -310,7 +314,7 @@ class ApplyBatchImpl:
             m,
             *(list(self.columns.values()) + list(self.table_args)),
             schema=schema,
-            **self.table_kwargs
+            **self.table_kwargs,
         )
 
         for const in (
@@ -352,14 +356,32 @@ class ApplyBatchImpl:
     def _gather_indexes_from_both_tables(self) -> List["Index"]:
         assert self.new_table is not None
         idx: List[Index] = []
-        idx.extend(self.indexes.values())
+
+        for idx_existing in self.indexes.values():
+            # this is a lift-and-move from Table.to_metadata
+
+            if idx_existing._column_flag:  # type: ignore
+                continue
+
+            idx_copy = Index(
+                idx_existing.name,
+                unique=idx_existing.unique,
+                *[
+                    _copy_expression(expr, self.new_table)
+                    for expr in _idx_table_bound_expressions(idx_existing)
+                ],
+                _table=self.new_table,
+                **idx_existing.kwargs,
+            )
+            idx.append(idx_copy)
+
         for index in self.new_indexes.values():
             idx.append(
                 Index(
                     index.name,
                     unique=index.unique,
                     *[self.new_table.c[col] for col in index.columns.keys()],
-                    **index.kwargs
+                    **index.kwargs,
                 )
             )
         return idx
@@ -400,7 +422,7 @@ class ApplyBatchImpl:
                             for elem in constraint.elements
                         ]
                     ],
-                    schema=referent_schema
+                    schema=referent_schema,
                 )
 
     def _create(self, op_impl: "DefaultImpl") -> None:
@@ -452,7 +474,7 @@ class ApplyBatchImpl:
         type_: Optional["TypeEngine"] = None,
         autoincrement: None = None,
         comment: Union[str, "Literal[False]"] = False,
-        **kw
+        **kw,
     ) -> None:
         existing = self.columns[column_name]
         existing_transfer: Dict[str, Any] = self.column_transfers[column_name]
@@ -462,13 +484,21 @@ class ApplyBatchImpl:
             existing.name = name
             existing_transfer["name"] = name
 
-            # pop named constraints for Boolean/Enum for rename
-            if (
-                "existing_type" in kw
-                and isinstance(kw["existing_type"], SchemaEventTarget)
-                and kw["existing_type"].name
-            ):
-                self.named_constraints.pop(kw["existing_type"].name, None)
+            existing_type = kw.get("existing_type", None)
+            if existing_type:
+                resolved_existing_type = _resolve_for_variant(
+                    kw["existing_type"], self.impl.dialect
+                )
+
+                # pop named constraints for Boolean/Enum for rename
+                if (
+                    isinstance(resolved_existing_type, SchemaEventTarget)
+                    and resolved_existing_type.name  # type:ignore[attr-defined]  # noqa E501
+                ):
+                    self.named_constraints.pop(
+                        resolved_existing_type.name,  # type:ignore[attr-defined]  # noqa E501
+                        None,
+                    )
 
         if type_ is not None:
             type_ = sqltypes.to_instance(type_)
@@ -565,7 +595,7 @@ class ApplyBatchImpl:
         column: "Column",
         insert_before: Optional[str] = None,
         insert_after: Optional[str] = None,
-        **kw
+        **kw,
     ) -> None:
         self._setup_dependencies_for_add_column(
             column.name, insert_before, insert_after
@@ -590,9 +620,11 @@ class ApplyBatchImpl:
         if (
             "existing_type" in kw
             and isinstance(kw["existing_type"], SchemaEventTarget)
-            and kw["existing_type"].name
+            and kw["existing_type"].name  # type:ignore[attr-defined]
         ):
-            self.named_constraints.pop(kw["existing_type"].name, None)
+            self.named_constraints.pop(
+                kw["existing_type"].name, None  # type:ignore[attr-defined]
+            )
 
     def create_column_comment(self, column):
         """the batch table creation function will issue create_column_comment
@@ -636,7 +668,8 @@ class ApplyBatchImpl:
                     if col_const.name == const.name:
                         self.columns[col.name].constraints.remove(col_const)
             else:
-                const = self.named_constraints.pop(cast(str, const.name))
+                assert const.name
+                const = self.named_constraints.pop(const.name)
         except KeyError:
             if _is_type_bound(const):
                 # type-bound constraints are only included in the new
