@@ -39,6 +39,7 @@ from alembic.testing import mock
 from alembic.testing import TestBase
 from alembic.testing.fixtures import op_fixture
 from alembic.util import exc as alembic_exc
+from alembic.util.sqla_compat import _safe_commit_connection_transaction
 from alembic.util.sqla_compat import _select
 from alembic.util.sqla_compat import has_computed
 from alembic.util.sqla_compat import has_identity
@@ -280,16 +281,20 @@ class BatchApplyTest(TestBase):
         create_stmt = re.sub(r"[\n\t]", "", create_stmt)
 
         idx_stmt = ""
-        for idx in impl.indexes.values():
+
+        # create indexes; these should be created in terms of the
+        # final table name
+        impl.new_table.name = impl.table.name
+
+        for idx in impl._gather_indexes_from_both_tables():
             idx_stmt += str(CreateIndex(idx).compile(dialect=context.dialect))
-        for idx in impl.new_indexes.values():
-            impl.new_table.name = impl.table.name
-            idx_stmt += str(CreateIndex(idx).compile(dialect=context.dialect))
-            impl.new_table.name = ApplyBatchImpl._calc_temp_name(
-                impl.table.name
-            )
+
         idx_stmt = re.sub(r"[\n\t]", "", idx_stmt)
 
+        # revert new table name to the temp name, assertions below
+        # are looking for the temp name
+        impl.new_table.name = ApplyBatchImpl._calc_temp_name(impl.table.name)
+
         if ddl_contains:
             assert ddl_contains in create_stmt + idx_stmt
         if ddl_not_contains:
@@ -356,6 +361,20 @@ class BatchApplyTest(TestBase):
         new_table = self._assert_impl(impl)
         eq_(new_table.c.x.name, "q")
 
+    def test_rename_col_w_index(self):
+        impl = self._ix_fixture()
+        impl.alter_column("tname", "y", name="y2")
+        new_table = self._assert_impl(
+            impl, ddl_contains="CREATE INDEX ix1 ON tname (y2)"
+        )
+        eq_(new_table.c.y.name, "y2")
+
+    def test_rename_col_w_uq(self):
+        impl = self._uq_fixture()
+        impl.alter_column("tname", "y", name="y2")
+        new_table = self._assert_impl(impl, ddl_contains="UNIQUE (y2)")
+        eq_(new_table.c.y.name, "y2")
+
     def test_alter_column_comment(self):
         impl = self._simple_fixture()
         impl.alter_column("tname", "x", comment="some comment")
@@ -1137,6 +1156,30 @@ class CopyFromTest(TestBase):
             "ALTER TABLE _alembic_tmp_foo RENAME TO foo",
         )
 
+    def test_change_name_from_existing_variant_type(self):
+        """test #982"""
+        context = self._fixture()
+        self.table.append_column(
+            Column("y", Text().with_variant(Text(10000), "mysql"))
+        )
+
+        with self.op.batch_alter_table(
+            "foo", copy_from=self.table
+        ) as batch_op:
+            batch_op.alter_column(
+                column_name="y",
+                new_column_name="q",
+                existing_type=Text().with_variant(Text(10000), "mysql"),
+            )
+        context.assert_(
+            "CREATE TABLE _alembic_tmp_foo (id INTEGER NOT NULL, "
+            "data VARCHAR(50), x INTEGER, q TEXT, PRIMARY KEY (id))",
+            "INSERT INTO _alembic_tmp_foo (id, data, x, q) "
+            "SELECT foo.id, foo.data, foo.x, foo.y FROM foo",
+            "DROP TABLE foo",
+            "ALTER TABLE _alembic_tmp_foo RENAME TO foo",
+        )
+
     def test_change_type_to_schematype(self):
         context = self._fixture()
         self.table.append_column(Column("y", Integer))
@@ -1282,6 +1325,20 @@ class BatchRoundTripTest(TestBase):
         context = MigrationContext.configure(self.conn)
         self.op = Operations(context)
 
+    def tearDown(self):
+        # why commit?  because SQLite has inconsistent treatment
+        # of transactional DDL. A test that runs CREATE TABLE and then
+        # ALTER TABLE to change the name of that table, will end up
+        # committing the CREATE TABLE but not the ALTER. As batch mode
+        # does this with a temp table name that's not even in the
+        # metadata collection, we don't have an explicit drop for it
+        # (though we could do that too).  calling commit means the
+        # ALTER will go through and the drop_all() will then catch it.
+        _safe_commit_connection_transaction(self.conn)
+        with self.conn.begin():
+            self.metadata.drop_all(self.conn)
+        self.conn.close()
+
     @contextmanager
     def _sqlite_referential_integrity(self):
         self.conn.exec_driver_sql("PRAGMA foreign_keys=ON")
@@ -1385,7 +1442,7 @@ class BatchRoundTripTest(TestBase):
                 type_=Integer,
                 existing_type=Boolean(create_constraint=True, name="ck1"),
             )
-        insp = inspect(config.db)
+        insp = inspect(self.conn)
 
         eq_(
             [
@@ -1440,7 +1497,7 @@ class BatchRoundTripTest(TestBase):
             batch_op.drop_column(
                 "x", existing_type=Boolean(create_constraint=True, name="ck1")
             )
-        insp = inspect(config.db)
+        insp = inspect(self.conn)
 
         assert "x" not in (c["name"] for c in insp.get_columns("hasbool"))
 
@@ -1450,7 +1507,7 @@ class BatchRoundTripTest(TestBase):
             batch_op.alter_column(
                 "x", type_=Boolean(create_constraint=True, name="ck1")
             )
-        insp = inspect(config.db)
+        insp = inspect(self.conn)
 
         if exclusions.against(config, "sqlite"):
             eq_(
@@ -1471,14 +1528,6 @@ class BatchRoundTripTest(TestBase):
                 [Integer],
             )
 
-    def tearDown(self):
-        in_t = getattr(self.conn, "in_transaction", lambda: False)
-        if in_t():
-            self.conn.rollback()
-        with self.conn.begin():
-            self.metadata.drop_all(self.conn)
-        self.conn.close()
-
     def _assert_data(self, data, tablename="foo"):
         res = self.conn.execute(text("select * from %s" % tablename))
         if sqla_14:
@@ -1492,7 +1541,7 @@ class BatchRoundTripTest(TestBase):
             batch_op.alter_column("data", type_=String(30))
             batch_op.create_index("ix_data", ["data"])
 
-        insp = inspect(config.db)
+        insp = inspect(self.conn)
         eq_(
             set(
                 (ix["name"], tuple(ix["column_names"]))
@@ -1661,9 +1710,10 @@ class BatchRoundTripTest(TestBase):
         pk_const = inspect(self.conn).get_pk_constraint("foo")
         eq_(pk_const["constrained_columns"], ["id"])
 
-    def test_add_pk_constraint(self):
+    @testing.combinations(("always",), ("auto",), argnames="recreate")
+    def test_add_pk_constraint(self, recreate):
         self._no_pk_fixture()
-        with self.op.batch_alter_table("nopk", recreate="always") as batch_op:
+        with self.op.batch_alter_table("nopk", recreate=recreate) as batch_op:
             batch_op.create_primary_key("newpk", ["a", "b"])
 
         pk_const = inspect(self.conn).get_pk_constraint("nopk")
@@ -1671,9 +1721,10 @@ class BatchRoundTripTest(TestBase):
             eq_(pk_const["name"], "newpk")
         eq_(pk_const["constrained_columns"], ["a", "b"])
 
+    @testing.combinations(("always",), ("auto",), argnames="recreate")
     @config.requirements.check_constraint_reflection
-    def test_add_ck_constraint(self):
-        with self.op.batch_alter_table("foo", recreate="always") as batch_op:
+    def test_add_ck_constraint(self, recreate):
+        with self.op.batch_alter_table("foo", recreate=recreate) as batch_op:
             batch_op.create_check_constraint("newck", text("x > 0"))
 
         ck_consts = inspect(self.conn).get_check_constraints("foo")
@@ -1682,12 +1733,13 @@ class BatchRoundTripTest(TestBase):
         )
         eq_(ck_consts, [{"sqltext": "x > 0", "name": "newck"}])
 
+    @testing.combinations(("always",), ("auto",), argnames="recreate")
     @config.requirements.check_constraint_reflection
-    def test_drop_ck_constraint(self):
+    def test_drop_ck_constraint(self, recreate):
         self._ck_constraint_fixture()
 
         with self.op.batch_alter_table(
-            "ck_table", recreate="always"
+            "ck_table", recreate=recreate
         ) as batch_op:
             batch_op.drop_constraint("ck", "check")
 
@@ -1731,11 +1783,41 @@ class BatchRoundTripTest(TestBase):
         )
 
     def _assert_table_comment(self, tname, comment):
-        insp = inspect(config.db)
+        insp = inspect(self.conn)
 
         tcomment = insp.get_table_comment(tname)
         eq_(tcomment, {"text": comment})
 
+    @testing.combinations(("always",), ("auto",), argnames="recreate")
+    def test_add_uq(self, recreate):
+        with self.op.batch_alter_table("foo", recreate=recreate) as batch_op:
+            batch_op.create_unique_constraint("newuk", ["x"])
+
+        uq_consts = inspect(self.conn).get_unique_constraints("foo")
+        eq_(
+            [
+                {"name": uc["name"], "column_names": uc["column_names"]}
+                for uc in uq_consts
+            ],
+            [{"name": "newuk", "column_names": ["x"]}],
+        )
+
+    @testing.combinations(("always",), ("auto",), argnames="recreate")
+    def test_add_uq_plus_col(self, recreate):
+        with self.op.batch_alter_table("foo", recreate=recreate) as batch_op:
+            batch_op.add_column(Column("y", Integer))
+            batch_op.create_unique_constraint("newuk", ["x", "y"])
+
+        uq_consts = inspect(self.conn).get_unique_constraints("foo")
+
+        eq_(
+            [
+                {"name": uc["name"], "column_names": uc["column_names"]}
+                for uc in uq_consts
+            ],
+            [{"name": "newuk", "column_names": ["x", "y"]}],
+        )
+
     @config.requirements.comments
     def test_add_table_comment(self):
         with self.op.batch_alter_table("foo") as batch_op:
@@ -1761,7 +1843,7 @@ class BatchRoundTripTest(TestBase):
         self._assert_table_comment("foo", None)
 
     def _assert_column_comment(self, tname, cname, comment):
-        insp = inspect(config.db)
+        insp = inspect(self.conn)
 
         cols = {col["name"]: col for col in insp.get_columns(tname)}
         eq_(cols[cname]["comment"], comment)
@@ -2004,7 +2086,7 @@ class BatchRoundTripTest(TestBase):
             ]
         )
         eq_(
-            [col["name"] for col in inspect(config.db).get_columns("foo")],
+            [col["name"] for col in inspect(self.conn).get_columns("foo")],
             ["id", "data", "x", "data2"],
         )
 
@@ -2030,7 +2112,7 @@ class BatchRoundTripTest(TestBase):
             ]
         )
         eq_(
-            [col["name"] for col in inspect(config.db).get_columns("foo")],
+            [col["name"] for col in inspect(self.conn).get_columns("foo")],
             ["id", "data", "x", "data2"],
         )
 
@@ -2051,7 +2133,7 @@ class BatchRoundTripTest(TestBase):
             tablename="nopk",
         )
         eq_(
-            [col["name"] for col in inspect(config.db).get_columns("foo")],
+            [col["name"] for col in inspect(self.conn).get_columns("foo")],
             ["id", "data", "x"],
         )
 
@@ -2071,7 +2153,7 @@ class BatchRoundTripTest(TestBase):
             ]
         )
         eq_(
-            [col["name"] for col in inspect(config.db).get_columns("foo")],
+            [col["name"] for col in inspect(self.conn).get_columns("foo")],
             ["id", "data2", "data", "x"],
         )
 
@@ -2091,7 +2173,7 @@ class BatchRoundTripTest(TestBase):
             ]
         )
         eq_(
-            [col["name"] for col in inspect(config.db).get_columns("foo")],
+            [col["name"] for col in inspect(self.conn).get_columns("foo")],
             ["id", "data", "data2", "x"],
         )
 
@@ -2125,12 +2207,12 @@ class BatchRoundTripTest(TestBase):
             ]
         )
         eq_(
-            [col["name"] for col in inspect(config.db).get_columns("foo")],
+            [col["name"] for col in inspect(self.conn).get_columns("foo")],
             ["id", "data", "x", "data2"],
         )
 
     def test_create_drop_index(self):
-        insp = inspect(config.db)
+        insp = inspect(self.conn)
         eq_(insp.get_indexes("foo"), [])
 
         with self.op.batch_alter_table("foo", recreate="always") as batch_op:
@@ -2145,8 +2227,7 @@ class BatchRoundTripTest(TestBase):
                 {"id": 5, "data": "d5", "x": 9},
             ]
         )
-
-        insp = inspect(config.db)
+        insp = inspect(self.conn)
         eq_(
             [
                 dict(
@@ -2162,7 +2243,7 @@ class BatchRoundTripTest(TestBase):
         with self.op.batch_alter_table("foo", recreate="always") as batch_op:
             batch_op.drop_index("ix_data")
 
-        insp = inspect(config.db)
+        insp = inspect(self.conn)
         eq_(insp.get_indexes("foo"), [])
 
 
@@ -2283,7 +2364,7 @@ class BatchRoundTripPostgresqlTest(BatchRoundTripTest):
         ) as batch_op:
             batch_op.add_column(Column("data", Integer))
 
-        insp = inspect(config.db)
+        insp = inspect(self.conn)
 
         eq_(
             [
